# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the Rectified Linear Unit (ReLU) activation function"""

import numpy as np
from . import ActivationFunction


class ReLU(ActivationFunction):
    """Rectified Linear Unit (ReLU) activation function.

    The ReLU activation function is the most common activation function in deep
    learning. It is a linear function that is bounded above by 0. One of the positive
    features of the ReLU activation function is that it is fast to compute and that helps
    avoid vanishing gradients.
    """

    def __init__(self):
        super().__init__(name='relu')

    @staticmethod
    def forward(x: np.ndarray) -> np.ndarray:
        """Computes the forward pass of the ReLU activation function."""

        # (x > 0) returns a boolean array, where True represents the elements of x
        # that are greater than 0. True is cast to 1.0 and False is cast to 0.0 when
        # the array is multiplied by x.
        return x * (x > 0)

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        """Computes the gradient of the ReLU activation function."""

        # z is the input to the activation function.
        # (z > 0) returns a boolean array, where True represents the elements of z
        # that are greater than 0. True is cast to 1.0 and False is cast to 0.0 when
        # the array is multiplied by z.
        return 1. * (z > 0)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the softmax activation function"""

import numpy as np
from .activation_function import ActivationFunction


class Softmax(ActivationFunction):
    """Softmax activation function.

    The softmax activation function is defined as:
    f(x) = exp(x) / sum(exp(x))

    The softmax activation function is a special case of the exponential
    activation function where the output is constrained to be between 0 and 1. It
    is often used as the last activation function in a network to ensure that
    the output is a probability distribution.

    """

    def __init__(self):
        super().__init__(name="softmax")

    @staticmethod
    def forward(x: np.ndarray) -> np.ndarray:
        """Compute the softmax of the input."""

        # Softmax is prone to overflow, so we use the following trick, extracted from:
        # https://stackoverflow.com/questions/42599498/numercially-stable-softmax

        # We subtract the max of each row to avoid overflow, thanks to the property that
        # softmax(x) = softmax(x + c) for all c.
        exp_x = np.exp(x - np.max(x, axis=1, keepdims=True))
        return exp_x / np.sum(exp_x, axis=1, keepdims=True)

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        """
        Compute the gradient of the softmax (d_a/d_z) of the input.
        Args:
            z: The input to the softmax function. It has shape (batch_size, num_classes).
        """
        # code adapted from:
        # https://stackoverflow.com/questions/36279904/softmax-derivative-in-numpy-approaches-0-implementation
        softmax = Softmax.forward(z)
        jacobian = - softmax[..., None] * softmax[:, None, :]  # off-diagonal Jacobian
        iy, ix = np.diag_indices_from(jacobian[0])
        jacobian[:, iy, ix] = softmax * (1. - softmax)  # diagonal

        # The code above is equivalent to the following, but is much faster (2x)*:
        # jacobian = np.empty((outputs.shape[0], outputs.shape[1], outputs.shape[1]))
        # softmax = Softmax.forward(outputs)
        # for m in range(outputs.shape[0]):
        #     for i in range(outputs.shape[1]):
        #         for j in range(outputs.shape[1]):
        #             jacobian[m, i, j] = softmax[m, i] * ((i == j) - softmax[m, j])
        #
        # *tested in the mnist

        return jacobian



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Home of the base activation function class."""

import numpy as np

from abc import ABC, abstractmethod


class ActivationFunction(ABC):
    """Base class for activation functions.

    An activation function takes a tensor as input and returns a tensor of the
    same shape. The activation function is usually applied element-wise.
    """

    def __init__(self, name):
        self.name = name

    # Methods
    # -------------------------------------------------------------------------

    @staticmethod
    @abstractmethod
    def forward(x: np.ndarray) -> np.ndarray:
        """Forward pass of the activation function.

        Args:
            x: Input of the activation function.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        """Compute the gradient of the activation function.

        Args:
            z: Input of the activation function.
        """
        raise NotImplementedError

    def __call__(self, x):
        return self.forward(x)

    def __str__(self):
        return self.name



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains CategoricalCrossentropy class."""

import numpy as np
from .loss_function import LossFunction


class CategoricalCrossentropy(LossFunction):
    """Cross entropy loss function.

    A loss function used in order to classify 2 or more categories. This class is commonly used to classify more than 2
    categories, such that if is not, binary cross-entropy is used.
    """

    def __init__(self, name="cross_entropy"):
        super(CategoricalCrossentropy, self).__init__(name)

    @staticmethod
    def compute_loss(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        """
        Args:
            y_true: the expected distribution of probabilities as a one-hot vector
            y_pred: the predicted distribution of probabilities

        Returns:
            A numpy array with just a single element
        """
        samples = len(y_pred)

        y_pred_clipped = np.clip(y_pred, 1e-7, 1 - 1e-7)

        correct_confidences = np.sum(y_pred_clipped * y_true, axis=1)

        negative_log_likelihoods = -np.log(correct_confidences)
        return np.mean(negative_log_likelihoods)

    @staticmethod
    def gradient(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        """

        Args:
            y_true: the expected distribution of probabilities as a one-hot vector
            y_pred: the predicted distribution of probabilities

        Returns:
            A numpy array with dimensions (samples, classes)
        """

        samples = len(y_pred)

        y_pred_clipped = np.clip(y_pred, 1e-7, 1 - 1e-7)

        gradients = -y_true / y_pred_clipped

        return gradients / samples



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the get_loss function."""

from . import MAE, MSE, BinaryCrossentropy, CategoricalCrossentropy, LossFunction


def get_loss_function(loss_name: str) -> LossFunction:
    """Returns the loss function corresponding to the loss name.

    Raises:
        ValueError: If the loss name is not recognized.
    """

    if loss_name == 'mae':
        return MAE()
    elif loss_name == 'mse':
        return MSE()
    elif loss_name == 'binary_crossentropy':
        return BinaryCrossentropy()
    elif loss_name == 'categorical_crossentropy':
        return CategoricalCrossentropy()
    else:
        raise ValueError(f'Unknown loss function: {loss_name}')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the convolutional layers classes (only the 2D convolution has been implemented by now)."""

import numpy as np
from typing import Tuple, Union
from .layer import Layer
from dlfs.activation_functions import ActivationFunction
from dlfs.convolutioners import Convolutioner, get_convolution


class Conv2D(Layer):
    """A 2D convolution layer.

    Conv2D is analogous to the Dense layer. They differ in the fact that Conv2D takes into account
    the spatial location. It also concerns some other parameters such as stride, kernel size, padding,
    etc; which are essential characteristics in order to carry out a convolution.

    Technical details (assuming data_format='channels_last'):
        - The input is a 4D tensor (batch, height, width, number of input channels).
        - The output is a 4D tensor (batch, height, width, n_filters). (see _get_output_shape for more details)
        - Each kernel is a 3D tensor (kernel_height, kernel_width, number of input channels).
        - The weights are a 4D tensor (kernel_height, kernel_width, number of input channels, n_filters).
        - The bias is a 1D tensor (n_filters,).
        - The layer performs a '3D' convolution on the input tensor for each kernel.



    Args:
        kernel_size (tuple): tuple of 2 integers, specifying the height and width of the 2D convolution window.
        n_filters (int): Number of filters/kernels.
        stride (tuple or int): specifying the strides of the convolution along the height and width.
            Can be a single integer to specify the same value for all spatial dimensions
        padding (bool, tuple or int): If True, add padding to the input so that the output has the same shape as the
            input (assuming stride = 1). If padding is a tuple of two integers, this defines the amount of padding
            to add to the top, bottom, left and right of the input. If padding is an integer, this number of zeros
            is added to the input on both sides.
        activation (ActivationFunction): Activation function
        use_bias (bool): Whether to use bias
        convolution_type (str): convolution mode. Can be 'winograd', 'direct' or 'patch'.
        name (str): Name of the layer
        input_shape (tuple): shape of the input
        weights_init (str): Initialization method for the weights
        bias_init (str): Initialization method for the bias

    Attributes:
        n_filters (int): Number of filters
        kernel_size (tuple): tuple of 2 integers, specifying the height and width of the 2D convolution window.
        stride (tuple): specifying the strides of the convolution along the height and width.
        padding (tuple): tuple of 2 integers, specifying the padding of the convolution along the height and width.
            This values are computed from the stride and kernel size in order to ensure that the output has the same
            shape as the input (if padding = True).
        activation (ActivationFunction): Activation function
        use_bias (bool): Whether to use bias.
        convolution_type (str): convolution mode. Recommended to be 'winograd'.
        name (str): Name of the layer
        blocksize (Tuple[int, int]): the size of the block, only used with Winograd.
        data_format (str): the data format of the input. Can be 'channels_last' or 'channels_first'.

    Raises:
        ValueError: If using padding and stride != 1.
    """

    def __init__(self,
                 n_filters: int,
                 kernel_size: Union[Tuple[int, int], int],
                 stride: Union[Tuple[int, int], int] = (1, 1),
                 padding: Union[bool, tuple, int] = False,
                 activation: str = None,
                 use_bias: bool = True,
                 convolution_type: str = "simple",
                 name: str = "Conv2D",
                 input_shape: tuple = None,
                 weights_init: str = "glorot_uniform",
                 bias_init: str = "zeros",
                 blocksize: Tuple[int, int] = None,
                 data_format: str = "channels_last"):

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if isinstance(stride, int):
            stride = (stride, stride)

        if n_filters < 0:
            raise ValueError("The number of filters must be greater than 0")

        if len(kernel_size) != 2:
            raise ValueError("The kernel size should be a tuple of two integers")

        if kernel_size[0] <= 0 or kernel_size[1] <= 0:
            raise ValueError("The kernel size should be greater than 0")

        input_shape = None if input_shape is None else (None, *input_shape)
        output_shape = None
        if isinstance(padding, bool):
            padding = (0, 0) if padding is False else (kernel_size[0] // 2, kernel_size[1] // 2)
        else:
            padding = (padding, padding) if isinstance(padding, int) else padding

        # get the output shape if the input shape is known
        if input_shape is not None:
            output_height = (input_shape[1] - kernel_size[0] + 2 * padding[0]) // stride[0] + 1
            output_width = (input_shape[2] - kernel_size[1] + 2 * padding[1]) // stride[1] + 1
            output_shape = (None, output_height, output_width, n_filters)

        super().__init__(input_shape=input_shape,
                         output_shape=output_shape,
                         activation=activation,
                         name=name)

        self.kernel_size = kernel_size
        self.n_filters = n_filters
        self.stride = stride
        self.padding = padding
        self.convolution_type = convolution_type
        self.blocksize = blocksize
        self.use_bias = use_bias
        self.weights_init = weights_init
        self.bias_init = bias_init
        self.forward_conv = None
        self.backward_conv = None
        self.update_conv = None
        self.data_format = data_format
        self._batch_count = 0  # needed to preprocess the data in the winograd algorithm

    def initialize(self, input_shape: tuple, weights: np.ndarray = None, bias: np.ndarray = None):
        """Initializes the layer.

        If weights and bias are not provided, they are initialized using the specified initialization method.

        Args:
            input_shape (tuple): input shape of the layer, it has the form (n_samples, height, width, n_channels)
            weights (np.ndarray): weights of the layer (optional, recommended to be None).
            bias (np.ndarray): bias of the layer (optional, recommended to be None).

        Raises:
            ValueError: if the input shape is not valid.
            ValueError: if the weights and bias are not of the shape:
                (input_shape[3], self.kernel_size[0], self.kernel_size[1], self.n_filters),
                (self.n_filters,), respectively.
            ValueError: if trying to set bias and `self.use_bias` is False.
        """

        # check if the input shape is correct
        if len(input_shape) != 4:
            raise ValueError("The input shape should be a tuple of four integers: "
                             "(n_samples, height, width, n_channels)")

        self.input_shape = input_shape
        self.forward_conv = get_convolution(self.convolution_type,
                                            input_shape[1:],
                                            self.kernel_size,
                                            self.padding,
                                            self.stride,
                                            self.blocksize)

        self.output_shape = self._get_output_shape()

        if self.data_format == "channels_first":
            weights_shape = (self.n_filters, input_shape[3], *self.kernel_size)
        else:
            weights_shape = (*self.kernel_size, input_shape[3], self.n_filters)

        # initialize weights
        if weights is not None:
            if weights.shape != weights_shape:
                raise ValueError(f"The shape of the weights should be "
                                 "(n_filters, kernel_height, kernel_width, n_channels_previous_layer). "
                                 f"Got {weights.shape}, expected {weights_shape}")
            self.weights = weights
        elif self.weights_init == "xavier":
            self.weights = np.random.normal(loc=0,
                                            scale=np.sqrt(
                                                1 / (input_shape[3] * self.kernel_size[0] * self.kernel_size[1])),
                                            size=weights_shape)
        elif self.weights_init == "zeros":
            self.weights = np.zeros(weights_shape)
        elif self.weights_init == "ones":
            self.weights = np.ones(weights_shape)
        elif self.weights_init == "uniform":
            self.weights = np.random.uniform(low=-1, high=1, size=weights_shape)
        elif self.weights_init == "normal":
            self.weights = np.random.normal(loc=0, scale=1, size=weights_shape)
        elif self.weights_init == "glorot_uniform":
            self.weights = np.random.uniform(low=-np.sqrt(6 / (input_shape[3] + self.n_filters)),
                                             high=np.sqrt(6 / (input_shape[3] + self.n_filters)),
                                             size=weights_shape)
        else:
            raise ValueError("Unknown weights initialization")

        if self.use_bias:
            bias_shape = (self.n_filters,)
            # initialize bias
            if bias is not None:
                if bias.shape != bias_shape:
                    raise ValueError(f"The shape of the bias should be "
                                     "(n_channels_current_layer). "
                                     f"Got {bias.shape}, expected {bias_shape}")
                self.bias = bias
            elif self.bias_init == "zeros":
                self.bias = np.zeros(bias_shape)
            elif self.bias_init == "ones":
                self.bias = np.ones(bias_shape)
            elif self.bias_init == "uniform":
                self.bias = np.random.uniform(low=-1, high=1, size=bias_shape)
            elif self.bias_init == "normal":
                self.bias = np.random.normal(loc=0, scale=1, size=bias_shape)
            else:
                raise ValueError("Unknown bias initialization")
        else:
            if bias is not None:
                raise ValueError("The bias should be None if the layer is not using bias")

        # INITIALIZE BACKWARD CONVOLUTION
        # compute the padding needed for the backward convolution
        padding_backward = ((self.kernel_size[0] - 1), (self.kernel_size[1] - 1))

        # initialize the backward convolution. This convolution is used to compute the gradient of the loss
        # with respect to the input.
        self.backward_conv = get_convolution(self.convolution_type,
                                             image_size=self.output_shape,  # delta.shape
                                             kernel_size=self.kernel_size,
                                             stride=self.stride,
                                             padding=padding_backward,
                                             data_format=self.data_format)

        # INITIALIZE UPDATE CONVOLUTION
        # This convolution is used to compute the gradient of the loss with respect to the weights.

        # The image size is the size of the output of the convolution but after applying the padding.
        if self.data_format == "channels_first":
            image_height, image_width = (self.output_shape[1], self.output_shape[2])
        else:
            image_height, image_width = (self.output_shape[2], self.output_shape[3])

        # compute the image size after applying the padding
        image_size = (image_height + 2 * padding_backward[0], image_width + 2 * padding_backward[1])

        self.update_conv = get_convolution(self.convolution_type,
                                           image_size=image_size,
                                           kernel_size=self.kernel_size,
                                           stride=self.stride,
                                           data_format=self.data_format)

        self.initialized = True

    def set_weights(self, weights: np.ndarray = None, bias: np.ndarray = None):
        """Sets the weights and bias of the layer.

        Args:
            weights (np.ndarray): weights of the layer.
            bias (np.ndarray): bias of the layer.

        Raises:
            ValueError: if the shape of the weights or bias is not correct.
            ValueError: if trying to set bias and `self.use_bias` is False.
        """
        if weights is not None:
            # check if the weights shape is correct
            if self.data_format == "channels_first":
                weights_shape = (self.n_filters, self.input_shape[1], self.kernel_size[0], self.kernel_size[1])
            else:
                weights_shape = (self.n_filters, self.kernel_size[0], self.kernel_size[1], self.input_shape[-1])
            if weights.shape != weights_shape:
                raise ValueError(f"The shape of the weights should be {weights_shape}, got {weights.shape}.")
            self.weights = weights

        if bias is not None:

            if not self.use_bias:
                raise ValueError("The layer does not use bias.")

            # check if the bias shape is correct
            bias_shape = (self.n_filters,)
            if bias.shape != bias_shape:
                raise ValueError(f"The shape of the bias should be "
                                 "(n_channels_current_layer). "
                                 f"Got {bias.shape}, expected {bias_shape}")

            self.bias = bias

    def _convolve(self,
                  x: np.ndarray,
                  kernel: np.ndarray,
                  convolutioner: Convolutioner,
                  data_format: str = "channels_last") -> np.ndarray:
        """Convolve a batch of multichannel images given a 'Convolutioner'

        Args:
            x (np.ndarray): batch of images.
            kernel (np.ndarray): kernel to convolve with.
            convolutioner (Convolutioner): convolutioner used to convolve the images.

        Returns:
            np.ndarray: batch of convolved images.
        """

        if data_format == "channels_last":
            outputs = np.array([convolutioner.convolve(x,
                                                       kernel[..., i],
                                                       batch_count=self._batch_count,
                                                       data_format=data_format) + self.bias[i]
                                for i in range(kernel.shape[-1])])

            # the shape of self.outputs.shape is (n_filters, batch_size, height, width). So we need to move the axis
            # to (batch_size, height, width, n_filters)
            outputs = np.moveaxis(outputs, 0, -1)
        else:  # data_format == "channels_first"
            outputs = np.array([convolutioner.convolve(x,
                                                       kernel[i],
                                                       batch_count=self._batch_count,
                                                       data_format=data_format) + self.bias[i]
                                for i in range(kernel.shape[0])])

            # the shape of self.outputs.shape is (n_filters, batch_size, height, width). So we need to move the axis
            # to (batch_size, n_filters, height, width)
            outputs = np.moveaxis(outputs, 0, 1)

        return outputs

    def forward(self, x: np.ndarray, training: bool = False) -> np.ndarray:
        """Forward pass

        Args:
            x: input data
            training: for compatibility with other layers

        Returns:
            output data
        """
        self.inputs = x
        self.outputs = self._convolve(x, self.weights, self.forward_conv, data_format=self.data_format)

        # update the batch count
        self._batch_count += 1
        # Take the modulo 2 to reduce memory usage. We only need to pass a batch count different from the previous one
        # to the update convolution, in order to indicate that the batch has changed. That's why we can do this.
        self._batch_count %= 2

        return self.outputs if self.activation is None else self.activation(self.outputs)

    def get_d_inputs(self, delta: np.ndarray) -> np.ndarray:
        """Returns the derivative of the cost function with respect to the input of the layer.

        Args:
            delta: derivative of the cost function with respect to the output of the layer.
        """
        # 180° rotation of the filters
        rotated_weights = np.rot90(self.weights, 2, axes=(1, 2)) if self.data_format == "channels_last" else \
            np.rot90(self.weights, 2, axes=(2, 3))
        a = 0
        a += 1

        # The current shape of the rotated weights is (height, width, n_input_channels, n_output_channels).
        # We need to exchange the number of input and output channels because we are in the backward pass.
        if self.data_format == "channels_last":
            rotated_weights = np.moveaxis(rotated_weights, 2, -1)
        else:
            rotated_weights = np.moveaxis(rotated_weights, 2, 0)
            rotated_weights = np.moveaxis(rotated_weights, -1, 0)

        d_inputs = self._convolve(delta, rotated_weights, self.backward_conv, data_format=self.data_format)
        return d_inputs

    def count_params(self) -> int:
        return self.weights.size + self.bias.size

    def _get_output_shape(self) -> Tuple[int, int, int, int]:
        """Returns the shape of the output of the layer, taking into account the already defined attributes.

        If the convolution mode is not 'winograd', the output shape could be computed as usual:
        batch_size = self.input_shape[0]
        output_height = (self.input_shape[1] - self.kernel_size[0] + 2 * self.padding[0]) // self.stride[0] + 1
        output_width = (self.input_shape[2] - self.kernel_size[1] + 2 * self.padding[1]) // self.stride[1] + 1

        return batch_size, output_height, output_width, self.n_filters

        However, if the convolution mode is 'winograd', the output shape is computed differently (see the
        `get_output_shape` method of the `WinogradConvolutioner` class for more details). For
         this reason we need to call the `get_output_shape` method of the `Convolution` class."""
        batch_size = self.input_shape[0]
        output_height, output_width = self.forward_conv.get_output_shape()
        return batch_size, output_height, output_width, self.n_filters

    def update(self, optimizer, delta: np.ndarray):
        """Updates the weights and biases of this layer.

        See `base class` for more details.

        Args:
            optimizer (Optimizer): optimizer used to update the weights and biases
            delta: delta of the loss with respect to the output of this layer
        """
        # check if the layer is initialized
        if not self.initialized:
            raise ValueError("The layer is not initialized")

        # Compute the gradient of the loss with respect to the weights and biases

        if self.data_format == "channels_last":
            # The self.inputs.shape is (batch_size, inputs_height, inputs_width, n_channels)
            # The delta.shape is (batch_size, inputs_height, inputs_width, n_filters)
            # The self.weights.shape is (kernel_height, kernel_width, n_channels, n_filters)

            # The gradient of the loss with respect to the weights is computed as:
            # d_w = convolve(inputs, delta)

            # We can not perform this operation directly because the convolution function expects the inputs to
            # have the shape (batch_size, inputs_height, inputs_width, n_channels) and the kernel (deltas) to have
            # the shape (kernel_height, kernel_width, n_channels). Thus, we have to use a list comprehension to
            # perform 'n_filters*batch_size' convolutioners and then sum the results and move the axes to the correct
            # positions.

            dw = np.zeros(self.weights.shape)
            for i, image in enumerate(self.inputs):
                for j in range(image.shape[2]):  # image.shape[2] = number of channels
                    for k in range(self.n_filters):
                        dw[:, :, j, k] += self.update_conv(image[:, :, j], delta[i, :, :, k], using_batches=False)

            dw = dw / self.inputs.shape[0]
            db = np.sum(delta, axis=(0, 1, 2))

        optimizer.update(self, (dw, db))

    def summary(self):
        return f"{self.name[:-1]}_{self.name[-1]}: output shape={self.output_shape}, n_params={self.count_params()}"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains mean square error loss function."""

import numpy as np
from .loss_function import LossFunction


class MSE(LossFunction):
    """Class that defines the Mean Squared Error loss function."""

    def __init__(self):
        super(MSE, self).__init__(name="mse")

    @staticmethod
    def compute_loss(y_true, y_pred):
        """
        Calculates the Mean Squared Error loss
        Args:
            y_true: expected outputs
            y_pred: predictions
        Returns:
            The mean squared error of the batch.

        """
        return np.sum(np.square(y_true - y_pred)) / (2 * y_true.shape[0])

    @staticmethod
    def gradient(y_true, y_pred):
        """Calculates the gradient of the Mean Squared Error loss with respect to the predictions

        Args:
            y_true: labels
            y_pred: predictions

        Returns:
            The gradient of the loss with respect to the predictions.
        """
        return y_pred - y_true



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the definition of the Dense layer class."""

import numpy as np
from .layer import Layer
from dlfs.optimizers import Optimizer


class Dense(Layer):
    """A fully-connected layer class.

    This is the most basic type of layer where the user states the number of neurons it is going to
    contain and optionally the activation function that is going to be needing. This kind of layers
    are utilize in both simple neuronal networks and convolutional networks. In the latter network,
    they are employed in the last layer to give out the output.

    The usage of this layer is very simple. The user just needs to specify the number of neurons
    and optionally the activation function. The activation function is optional and if not specified
    it will use a linear activation function.

    Args:
        n_neurons (int): The number of neurons that the layer will contain.
        activation (function): The activation function that the layer will use.
        name (str): The name of the layer.
        input_shape (tuple): The shape of the input. Required for the first layer.
        weights_init (str): The initialization method for the weights. Default is 'xavier'.
        bias_init (str): The initialization method for the bias. Default is 'zeros'.
    """

    def __init__(self,
                 n_neurons: int,
                 activation: str = None,
                 name: str = "Dense",
                 input_shape: tuple = None,
                 weights_init: str = "xavier",
                 bias_init: str = "zeros"):

        if n_neurons <= 0:
            raise ValueError("The number of neurons should be greater than 0")

        # In order to give an easier way to define the layer, we allow the user to define the input shape
        # without having to specify the batch size, that is why we set the batch size to None.
        input_shape = None if input_shape is None else (None, *input_shape)
        super(Dense, self).__init__(input_shape=input_shape,
                                    output_shape=(None, n_neurons),
                                    activation=activation,
                                    name=name)
        self.__n_neurons = n_neurons
        self.weights_init = weights_init  # recommended: xavier
        self.bias_init = bias_init  # recommended: zeros

    # Getters
    # ----------------------------------------------------------------------------------------------------

    @property
    def n_neurons(self):
        return self.__n_neurons

    # Methods
    # ----------------------------------------------------------------------------------------------------
    def initialize(self, input_shape: tuple, weights: np.ndarray = None, bias: np.ndarray = None):
        """Initialize the layer. This method is called by the model when the model is being compiled.

        Args:
            input_shape (tuple): input shape of the layer, it has the form (n_samples (None), n_features)
            weights (np.ndarray): weights of the layer (optional, recommended to be None).
                The weights has the shape (n_neurons_prev_layer, n_neurons_current_layer). Each column of the weights
                matrix represents a neuron and the values of the column are its weights.
            bias (np.ndarray): bias of the layer (optional, recommended to be None). The bias has the shape
                (1, n_neurons_current_layer).

        Raises:
            ValueError: If the input shape is not valid.
            ValueError: If the weights and bias do not match the expected shape.
            ValueError: If the weights or bias initialization method is not valid.
        """
        # check if the input shape is correct
        if len(input_shape) != 2:
            raise ValueError("The input shape is incorrect")

        self.input_shape = input_shape

        if weights is not None:
            self.weights = weights
        elif self.weights_init == "xavier":
            # The recommended initialization [https://www.deeplearning.ai/ai-notes/initialization/]
            self.weights = np.random.normal(loc=0,
                                            scale=np.sqrt(2 / (input_shape[1] + self.n_neurons)),
                                            size=(input_shape[1], self.n_neurons))
        elif self.weights_init == "zeros":
            self.weights = np.zeros((input_shape[1], self.n_neurons))
        elif self.weights_init == "ones":
            self.weights = np.ones((input_shape[1], self.n_neurons))
        elif self.weights_init == "normal":
            self.weights = np.random.normal(loc=0,
                                            scale=1,
                                            size=(input_shape[1], self.n_neurons))
        elif self.weights_init == "uniform":
            self.weights = np.random.uniform(low=-1,
                                             high=1,
                                             size=(input_shape[1], self.n_neurons))
        else:
            raise ValueError("The weights initializer is incorrect")

        if bias is not None:
            self.bias = bias
        elif self.bias_init == "zeros":
            # the recommended initialization [https://www.deeplearning.ai/ai-notes/initialization/]
            self.bias = np.zeros((1, self.n_neurons))
        elif self.bias_init == "ones":
            self.bias = np.ones((1, self.n_neurons))
        elif self.bias_init == "normal":
            self.bias = np.random.normal(loc=0,
                                         scale=1,
                                         size=(1, self.n_neurons))
        elif self.bias_init == "uniform":
            self.bias = np.random.uniform(low=-1,
                                          high=1,
                                          size=(1, self.n_neurons))
        else:
            raise ValueError("The bias initializer is incorrect")

        self.initialized = True

    def set_weights(self, weights: np.ndarray = None, bias: np.ndarray = None):
        """
        Set the weights and bias of the layer.

        Args:
            weights (np.ndarray): weights of the layer.
                The weights has the shape (n_neurons_prev_layer, n_neurons_current_layer). Each column of the weights
                matrix represents a neuron and the values of the column are its weights.
            bias (np.ndarray): bias of the layer. The bias has the shape (1, n_neurons_current_layer).
        """

        if weights is not None:
            # check the shape of the weights
            if weights.shape != (self.input_shape[1], self.n_neurons):
                raise ValueError(f"The shape of the weights is incorrect. Received {weights.shape}, expected "
                                 f"{(self.input_shape[1], self.n_neurons)}")
            self.weights = weights

        if bias is not None:
            # check the shape of the bias
            if bias.shape != (1, self.n_neurons):
                raise ValueError(f"The shape of the bias is incorrect. Received {bias.shape}, expected "
                                 f"{(1, self.n_neurons)}")
            self.bias = bias

    def forward(self, inputs, training: bool = False) -> np.ndarray:
        """Forward pass of the layer.

        It realizes the following operation:
            outputs = inputs @ weights + bias


        Args:
            inputs (np.ndarray): inputs of the layer
            training (bool): for compatibility with other layers

        Returns:
            np.ndarray: outputs of the layer
        """
        # check if the layer is initialized
        if not self.initialized:
            raise ValueError("The layer is not initialized")
        # check if the input shape is correct
        if inputs.shape[1:] != self.input_shape[1:]:
            raise ValueError("The input shape is incorrect")

        # check that the weights and bias are correct. This is needed since the weights and bias can be initialized
        # manually by the user
        if self.weights.shape[0] != self.input_shape[1]:
            raise ValueError("The weights must have the shape (n_features, n_neurons)")
        if self.bias.shape[1] != self.n_neurons:
            raise ValueError("The bias must have the shape (1, n_neurons)")

        # save the inputs
        self.inputs = inputs
        self.outputs = inputs @ self.weights + self.bias
        # check overflow
        if np.any(np.isnan(self.outputs)):
            raise OverflowError("An overflow occurred during the forward pass")

        return self.outputs if self.activation is None else self.activation(self.outputs)

    def get_d_inputs(self, delta: np.ndarray) -> np.ndarray:
        """
        Compute the gradient of the loss with respect to the inputs of the layer.

        Args:
            delta (np.ndarray): delta of the layer.

        Returns:
            The gradient of the loss with respect to the inputs of the layer.
        """

        # check if the layer is initialized
        if not self.initialized:
            raise ValueError("The layer is not initialized")

        # compute the gradient of the loss with respect to the inputs
        return delta @ self.weights.T

    def get_dz_da(self) -> np.ndarray:
        """
        Returns:
            The derivative of the output of this layer (i) with respect to the outputs of the next layer (i+1).
        """
        # check if the layer is initialized
        if not self.initialized:
            raise ValueError("The layer is not initialized")

        return self.weights.T

    def update(self, optimizer: Optimizer, delta: np.ndarray):
        """
        Update the weights and biases of the layer.

        Args:
            optimizer (Optimizer): optimizer used to update the weights and biases.
            delta (np.ndarray): delta of the current layer
        """

        # check if the layer is initialized
        if not self.initialized:
            raise ValueError("The layer is not initialized")

        d_weights = (self.inputs.T @ delta) / self.inputs.shape[0]
        d_bias = delta.sum(axis=0, keepdims=True) / self.inputs.shape[0]

        # check overflow
        if np.isnan(d_weights).any() or np.isnan(d_bias).any():
            raise OverflowError("The gradient is too large")

        optimizer.update(self, (d_weights, d_bias))

        # check overflow in the weights and biases
        if np.any(np.isnan(self.weights)) or np.any(np.isnan(self.bias)):
            raise OverflowError("An overflow occurred during the update")

    def count_params(self) -> int:
        """
        Count the number of parameters of the layer.

        Returns:
            int: number of parameters of the layer
        """
        return self.weights.size + self.bias.size

    def summary(self) -> str:
        """Summary of the layer.

        Returns:
            str: summary of the layer
        """
        return f"{self.name}: {self.__n_neurons} neurons\t output_shape={self.output_shape}\t " \
               f"n_params={self.weights.size + self.bias.size}"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Home of the metric base class."""

from abc import ABC, abstractmethod
import numpy as np


class Metric(ABC):
    """Base class for all metrics.

    A metric is a function that takes two arguments, a prediction and a label,
    and returns a scalar value.
    """

    def __init__(self, name):
        self.name = name

    @staticmethod
    @abstractmethod
    def compute_metric(y_true: np.ndarray, y_pred: np.ndarray) -> float:
        pass

    def __str__(self):
        return self.name

    def __call__(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        return self.compute_metric(y_true, y_pred)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the train test split method function"""

import numpy as np


def train_test_split(x: np.ndarray, y: np.ndarray, test_size: float = 0.3, random_state: int = None):
	"""Splits the dataset into train and test sets.

	Args:
		x: input data
		y: labels
		test_size: the proportion of the dataset to include in the test split
		random_state: the seed used by the random number generator

	Returns:
		x_train: training data
		y_train: training labels
		x_test: testing data
		y_test: testing labels
	"""
	train = x.shape[0] - int(x.shape[0] * test_size)
	arr = np.arange(x.shape[0])
	np.random.RandomState(seed=random_state).shuffle(arr)
	return x[arr[:train]], x[arr[train:]], y[arr[:train]], y[arr[train:]]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the get_optimizer function."""

from . import Optimizer, SGD, SGDMomentum


def get_optimizer(optimizer_name: str) -> Optimizer:
    """Get an optimizer by name.

    Args:
        optimizer_name (str): The name of the optimizer. Valid names are:
            - sgd
            - sgd_momentum
            - adam
            - adagrad

    Returns:
        The optimizer.
    """
    if optimizer_name == "sgd":
        return SGD()
    elif optimizer_name == "sgd_momentum":
        return SGDMomentum()
    else:
        raise ValueError(f"Unknown optimizer name: {optimizer_name}")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the sigmoid activation function"""

import numpy as np
from .activation_function import ActivationFunction


class Sigmoid(ActivationFunction):
    """Sigmoid activation function.

    The sigmoid activation function is defined as:
    f(x) = 1 / (1 + e^(-x))

    It returns a value between 0 and 1.
    """

    def __init__(self):
        super().__init__(name='sigmoid')

    @staticmethod
    def forward(x: np.ndarray) -> np.ndarray:
        """Computes the sigmoid activation function."""
        return 1 / (1 + np.exp(-x))

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        """Computes the gradient of the sigmoid activation function."""
        sigmoid = Sigmoid.forward(z)
        return sigmoid * (1 - sigmoid)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the WinogradConvolutioner class"""

import numpy as np
from typing import Union, Tuple, List
import tensorly
from skimage.util import view_as_blocks
from dlfs.convolutioners import Convolutioner


class WinogradConvolutioner(Convolutioner):
    """A convolutioner that uses the Winograd convolution algorithm.

    The Winograd algorithm reduces the number of multiplication operations by transforming the input feature map and
    executing a series of transformations on the filter. The result is a much faster convolution.

    This class is not ready for use yet.
    """

    def __init__(self,
                 image_size: Union[int, tuple],
                 kernel_size: Union[int, tuple],
                 padding: Union[int, tuple] = (0, 0),
                 stride: Union[int, tuple] = (1, 1),
                 blocksize: Tuple[int, int] = None,
                 data_format: str = 'channels_last'):

        super().__init__(image_size, kernel_size, padding, stride, data_format=data_format)

        if blocksize is None:
            self.blocksize = None
            self.set_block_size()
        else:
            self.blocksize = blocksize if isinstance(blocksize, tuple) else (blocksize, blocksize)
        self.__transformed_images = []
        self.__transformed_filter = None
        self.block_output_size = None
        self.get_output_shape()
        self.__y_t_matrices = []
        self.__image_transformers = []  # X
        self.__filter_transformers = []  # W
        self.__get_matrices()  # List of arrays
        self.batch_count = 0

    @staticmethod
    def convolve_multichannel(image: np.ndarray,
                              kernel: np.ndarray,
                              stride: Union[int, tuple] = (1, 1),
                              blocksize: Tuple[int, int] = None,
                              transformed_image: np.ndarray = None,
                              y_t_matrices: List[np.ndarray] = None,
                              image_transformers=None) -> np.ndarray:
        out = np.array([tensorly.tenalg.multi_mode_dot(i * kernel, y_t_matrices, list(range(3))) for i in transformed_image])
        print(out)

    @staticmethod
    def convolve_grayscale(image: np.ndarray,
                           kernel: np.ndarray,
                           stride: Union[int, tuple] = (1, 1),
                           blocksize: Tuple[int, int] = None,
                           transformed_image: np.ndarray = None,
                           y_t_matrix=None,
                           image_transformer=None) -> np.ndarray:
        out = np.array([y_t_matrix[0] @ i @ y_t_matrix[1].T for i in transformed_image])
        print(out)

    def __get_matrices(self):
        calculated = []
        for i, j in zip(self.block_output_size, self.image_size):
            if (i, j) in calculated:
                index = calculated.index((i, j))
                self.__y_t_matrices.append(self.__y_t_matrices[index])
                self.__image_transformers.append(self.__image_transformers[index])
                self.__filter_transformers.append(self.__filter_transformers[index])
            else:
                yt, x, w = WinogradConvolutioner.winograd_get_matrices(i, j)
                self.__y_t_matrices.append(yt)
                self.__image_transformers.append(x)
                self.__filter_transformers.append(w)
                calculated.append((i, j))

    def get_output_shape(self) -> Tuple[int, int]:
        """Returns the output shape of the convolution."""
        '''if not len(self.image_size) == len(self.kernel_size):
            raise ValueError('Image and kernel must have the same number of dimensions.')'''

        self.block_output_size = np.array([self.blocksize[i] + self.kernel_size[i] - 1 for i in range(len(self.blocksize))])
        num_of_blocks = np.array((self.image_size[1] / self.block_output_size[0], self.image_size[2] / self.block_output_size[1]))

        if not float(np.prod(num_of_blocks)).is_integer():
            raise ValueError(f'The image size must be divisible by the blocksize. {self.image_size} is not divisible '
                             f'by {self.block_output_size}.')

        output_size = self.block_output_size * num_of_blocks
        return tuple(output_size)

    @staticmethod
    def __transform_multichannel(block: np.ndarray, x: List[np.ndarray]):
        """Returns the image transform for a multichannel image.

        Args:
            block: The image to transform.
            x: The image transform matrices.
        """

        # if block.ndim == 2:
        #     return x[0].T @ block @ x[1]
        # elif block.ndim == 3:
        #     return tensorly.tenalg.multi_mode_dot(block, x, list(range(block.ndim)))

        print(block.shape)
        return np.array([tensorly.tenalg.multi_mode_dot(block[i, j, 0], x, list(range(3))) for j in range(block.shape[1])
                         for i in range(block.shape[0])])

    @staticmethod
    def __transform_grayscale(block: np.ndarray, x: List[np.ndarray]):
        """Returns the image transform for a multichannel image.

        Args:
            block: The image to transform.
            x: The image transform matrices.
        """

        # if block.ndim == 2:
        #     return x[0].T @ block @ x[1]
        # elif block.ndim == 3:
        #     return tensorly.tenalg.multi_mode_dot(block, x, list(range(block.ndim)))

        return np.array([x[0].T @ block[i, j] @ x[1] for j in range(block.shape[1]) for i in range(block.shape[0])])

    @staticmethod
    def __transform_filter(kernel: np.ndarray, m: List[np.ndarray]):
        """Returns the image transform for a multichannel image.

        Args:
            kernel: The filter to transform.
            m: The image transform matrices.
        """

        if kernel.ndim == 2:
            return m[0] @ kernel @ m[1].T
        elif kernel.ndim == 3:
            return tensorly.tenalg.multi_mode_dot(kernel, m, list(range(kernel.ndim)))

    def set_block_size(self):
        num_of_passes = (self.image_size[1] - self.kernel_size[0], self.image_size[2] - self.kernel_size[1])
        shape1 = -1
        shape2 = -1
        print(self.image_size)
        for i in range(self.kernel_size[0] + 1, self.image_size[1]):
            if num_of_passes[0] % i == 0:
                shape1 = i
                print(i)
                break
        for i in range(self.kernel_size[1] + 1, self.image_size[2]):
            if num_of_passes[1] % i == 0:
                shape2 = i
                print(num_of_passes, i)
                break
        self.blocksize = np.array((shape1, shape2))

    def convolve(self,
                 x: np.ndarray,
                 kernel: np.ndarray,
                 using_batches: bool = True,
                 data_format: str = 'channels_last',
                 **kwargs) -> np.ndarray:
        """Returns the convolution of the image and kernel.

        Args:
            x: The image to convolve.
            kernel: The kernel to convolve with.
            using_batches: Whether or not x is a batch of images.
            data_format: The data format of the image.
            **kwargs: Additional keyword arguments. Allowed keywords are:
                - 'batch_count': Used to know whether or not a preprocessing step is needed. If the batch count is
                    different from the last batch count or if it is not provided, we are dealing with a new batch,
                    so we need to preprocess the images and kernels.

        """

        # Add padding to the image if necessary
        if self.padding != (0, 0):
            x = self.pad_image(x, self.padding, data_format=data_format, using_batches=using_batches)

            self.__transformed_filter = WinogradConvolutioner.__transform_filter(kernel, self.__filter_transformers)

        if using_batches:
            # Move batch dimension to the last dimension
            x = np.moveaxis(x, 0, -1)

            if x.ndim == 4:

                # currently only supports data_format = 'channels_last'. So...
                if data_format != 'channels_last':
                    x = np.moveaxis(x, 0, 2)

                # get the blocks
                n_channels = x.shape[2]  # The only data format allowed is 'channels_last'

                # blocks contains 'batch_size' number of numpy arrays each one containing the
                # blocks of the image (views)
                print(x[...,0].shape, (*self.block_output_size, n_channels))
                blocks = [view_as_blocks(image, (*self.block_output_size, n_channels)) for image in x]

                if self.batch_count != kwargs.get('batch_count', None):
                    # blocks contains 'batch_size' number of numpy arrays each one containing the
                    # blocks of the image (views)
                    self.blocks = [view_as_blocks(x[..., i], (*self.blocksize, n_channels)) for i in range(x.shape[-1])]
                    self.batch_count = kwargs['batch_count']
                    self.__transformed_images = [
                        WinogradConvolutioner.__transform_multichannel(block, self.__image_transformers,)
                        for block in blocks]

                return np.array([self.convolve_multichannel(x[i],
                                                            self.__transformed_filter,
                                                            self.stride,
                                                            **kwargs,
                                                            transformed_image=self.__transformed_images[i],
                                                            y_t_matrices=self.__y_t_matrices)
                                 for i in range(x.shape[0])])
            elif x.ndim == 3:

                # get the blocks
                blocks = [view_as_blocks(image, tuple(self.block_output_size)) for image in x]
                self.__transformed_images = [WinogradConvolutioner.__transform_grayscale(block,
                                                                                            self.__image_transformers)
                                             for block in blocks]
                return np.array([self.convolve_grayscale(image, self.__transformed_filter, self.stride, **kwargs,
                                                         transformed_image=self.__transform_multichannel(x,
                                                                                                         self.__filter_transformers),
                                                         y_t_matrix=self.__y_t_matrices,
                                                         image_transformer=self.__image_transformers) for image in x])
        else:
            if x.ndim == 2:
                return self.convolve_grayscale(x, kernel, self.stride, **kwargs,
                                               transformed_image=self.__transform_multichannel(x,
                                                                                               self.__filter_transformers),
                                               y_t_matrix=self.__y_t_matrices,
                                               image_transformer=self.__image_transformers)
            elif x.ndim == 3:
                return self.convolve_multichannel(x, kernel, self.stride, **kwargs,
                                                  transformed_image=
                                                  self.__transform_multichannel(x, self.__filter_transformers),
                                                  y_t_matrices=self.__y_t_matrices,
                                                  image_transformers=self.__image_transformers)

        raise ValueError("Image must be 2D or 3D.")

    @staticmethod
    def vandermonde_matrix(b: int, points: List[Tuple[float, int]]) -> np.ndarray:
        """
        This is equivalent to V_{axb} on the document.
        This function returns a trimmed Vandermonde Matrix of size len(points) image b

            f_0^0 * g_0^b-1       f_0^1 * g_0^b-2       ... f_0^b-1 * g_0^0
            .
            .
            f_a-1^0 * g_a-1^b-1   f_a-1^1 * g_a-1^b-2   ... f_a-1^b-1 * g_a-1^0

        Arguments:
            b (int) : One of the dimensions of the matrix
            points (list[tuple[Union[int, float]]]) : List of homogeneous coordinates
        Returns:
            V (np.ndarray[Union[int, float]]) : The trimmed Vandermonde matrix
        """
        return np.array([[i[0] ** j * i[1] ** (b - j - 1) for j in range(b)] for i in points])

    @staticmethod
    def gen_points(num_points: int) -> List[Tuple[float, int]]:
        """Returns a list of homogeneous coordinates of size num_points

        This function generates homogeneous coordinates, starting from (0, 1), then, (1/2, 1), then (-1/2, 1),
        and the infinite point (0, 1).

            [(f_0, g_0), ..., (f_a-1, g_a-1)]

        Args:
            num_points (int): The desired number of points
        """
        points = [(0, 1)]
        if num_points % 2:
            points.extend([(y / 2, 1) for x in range(1, num_points // 2) for y in (x, -x)])
            points.extend([(num_points // 2 / 2, 1), (1, 0)])
        else:
            points.extend([(y / 2, 1) for x in range(1, num_points // 2) for y in (x, -x)])
            points.append((1, 0))
        return points

    @staticmethod
    def winograd_get_matrices(m: int, n: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        This is equivalent to F(m, n) on the paper.
        This function returns the matrices necessary for calculating the Winograd convolution given the dimension of the filter and the dimension of the output. With those, we can infer the size of the input.
        The usage of these matrices depends on the dimension of the convolution. If we denote image as the signal and w as the filter:
        - If it is 1D: F(m, n) = image * w = Y.T @ [(X.T @ image) * (W @ w)]
        - If it is 2D: F(m image i, n image j) = image * w = Y.T @ [(X.T @ image @ X') * (W @ w @ W'.T)] @ Y'
        In the latter case, if m = i and n = j, then Y = Y', X = X', W = W'. If not, then you would have to calculate both F(m, n) to get Y, X, W and F(i, j) to get Y', X', W'.

        Args:
            m (int): The size of the output
            n (int): The size of the filter

        Returns:
            A tuple with the aforementioned Y.T, X, W in this order.
        """
        num_points = m + n - 1
        points = WinogradConvolutioner.gen_points(num_points)
        y = WinogradConvolutioner.vandermonde_matrix(m, points)
        x = np.linalg.inv(WinogradConvolutioner.vandermonde_matrix(num_points, points))
        w = WinogradConvolutioner.vandermonde_matrix(n, points)
        return y.T, x, w



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Home of the base Convolutioner class."""

from abc import ABC, abstractmethod
import numpy as np
from typing import Union, Tuple
from patchify import patchify


class Convolutioner(ABC):
    """Abstract class for convolutioners.

    A convolutioner is a class that can be used to convolve a given image with
    a given kernel. It does not store any kernel or image data, but instead
    provides a method to convolve the image with the kernel. However, WinogradConvolutioner
    makes some pre-processing steps to make the convolution faster storing transformed matrices based on the kernel
    and the image dimensions. This is the reason the convolutioner takes this parameters as input. Currently,
    it does not support dilation.

    Args:
        image_size (tuple[int, int] or int): The size of the image to convolve. If an int is provided,
            it is assumed to be the size of the image along the first two dimensions.
        kernel_size (tuple[int, int] or int): The size of the kernel to convolve. If an int is provided,
            it is assumed to be the size of the kernel along the first two dimensions.
        padding (tuple[int, int] or int): The padding to apply to the image. If an int is provided,
            it is assumed to be the padding along the first two dimensions.
        stride (tuple[int, int] or int): The stride of the convolution. If an int is provided,
            it is assumed to be the stride along the first two dimensions.
        data_format (str): The data format of the image and kernel. Must be either 'channels_first' or 'channels_last'.


    Attributes:
        image_size: The size of the image to be convolved.
        kernel_size: The size of the kernel to be convolved with.
        padding: The amount of padding to be used.
        stride: The amount of stride to be used.
    """

    def __init__(self,
                 image_size: Union[int, tuple],
                 kernel_size: Union[int, tuple],
                 padding: Union[int, tuple] = (0, 0),
                 stride: Union[int, tuple] = (1, 1),
                 data_format: str = "channels_last"):

        self.image_size = image_size if isinstance(image_size, tuple) else (image_size, image_size)
        self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
        self.padding = padding if isinstance(padding, tuple) else (padding, padding)
        self.stride = (stride, stride) if isinstance(stride, int) else stride
        self.data_format = data_format

    @staticmethod
    @abstractmethod
    def convolve_grayscale(image: np.ndarray,
                           kernel: np.ndarray,
                           stride: Union[int, tuple] = (1, 1),
                           data_format: str = "channels_last",
                           **kwargs) -> np.ndarray:
        """Convolves a grayscale image with a grayscale kernel.

        Args:
            image (np.ndarray): The image to convolve.
            kernel (np.ndarray): The kernel to convolve with.
            stride (Union[int, tuple]): The stride to use.
            data_format (str): The data format of the image and kernel. Must be either 'channels_first' or
            'channels_last'.
            **kwargs: Additional keyword arguments.

        Returns:
            np.ndarray: The result of the convolution.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def convolve_multichannel(image: np.ndarray,
                              kernel: np.ndarray,
                              stride: Union[int, tuple] = (1, 1),
                              data_format: str = "channels_last",
                              **kwargs) -> np.ndarray:
        """Convolves a multichannel image with a 3D kernel.

        This video of DeepLearning.ai explains the process:
        https://www.youtube.com/watch?v=KTB_OFoAQcc&t=9s

        Note that the number of channels in the image and kernel must be equal.

        Args:
            image (np.ndarray): The image to convolve.
            kernel (np.ndarray): The kernel to convolve with.
            stride (Union[int, tuple]): The stride to use.
            data_format (str): The data format of the image and kernel. Must be either 'channels_first' or
            'channels_last'.
            **kwargs: Additional keyword arguments.

        Returns:
            np.ndarray: The result of the convolution.
        """
        raise NotImplementedError

    @staticmethod
    def get_patches(x: np.ndarray,
                    patch_size: Union[Tuple[int, int], Tuple[int, int, int]],
                    step: int = 1,
                    using_batches: bool = False) -> np.ndarray:
        """Returns the patches of the image.

        Args:
            x: The image to extract patches from. The image must be a numpy array and must be already
                padded (if necessary).
            patch_size: The size of the patches to extract.
            step: The step size between patches.
            using_batches: Whether to use batches or not.

        Returns:
            The patches of the image.
        """
        # we make use of the patchify library to extract the patches
        if using_batches:
            return np.array([patchify(image, patch_size, step) for image in x])
        else:
            return patchify(x, patch_size, step)

    def get_output_shape(self) -> Tuple[int, int]:
        """Returns the output shape of the convolution."""
        output_height = (self.image_size[0] - self.kernel_size[0] + 2 * self.padding[0]) // self.stride[0] + 1
        output_width = (self.image_size[1] - self.kernel_size[1] + 2 * self.padding[1]) // self.stride[1] + 1
        return output_height, output_width

    def convolve(self,
                 x: np.ndarray,
                 kernel: np.ndarray,
                 using_batches: bool = True,
                 data_format: str = "channels_last",
                 **kwargs) -> np.ndarray:
        """Convolves the image with the given kernel.

        This method is used as an abstraction layer for the convolution of grayscale and multichannel
        images. It calls the appropriate convolution method depending on the number of channels in the
        image.

        Args:
            x (np.ndarray): The image to convolve.
            kernel (np.ndarray): The kernel to convolve with.
            using_batches (bool): Whether if the 'x' argument is a batch of images or not.
            data_format (str): The data format of the image and kernel. Must be either 'channels_first' or
                'channels_last'.
            **kwargs: Additional keyword arguments.
        """

        # Add padding to the image if necessary
        if self.padding != (0, 0):
            x = self.pad_image(x, self.padding, self.data_format, using_batches)

        if using_batches:
            if x.ndim == 4:
                return np.array([self.convolve_multichannel(image, kernel, self.stride, data_format, **kwargs)
                                 for image in x])
            elif x.ndim == 3:
                return np.array([self.convolve_grayscale(image, kernel, self.stride, data_format, **kwargs)
                                 for image in x])
        else:
            if x.ndim == 2:
                return self.convolve_grayscale(x, kernel, self.stride, **kwargs)
            elif x.ndim == 3:
                return self.convolve_multichannel(x, kernel, self.stride, **kwargs)

        raise ValueError("Image must be 2D or 3D.")

    @staticmethod
    def pad_image(x: np.ndarray,
                  padding: Union[int, tuple],
                  data_format: str = "channels_last",
                  using_batches: bool = False) -> np.ndarray:

        padding = (padding, padding) if isinstance(padding, int) else padding

        # If data format is channels_last, then we need to convert the image to channels_first

        if not using_batches:
            if x.ndim == 3:  # A single multichannel image
                # If the data format is channels_last, then we need to convert the image to channels_first
                x = np.moveaxis(x, -1, 0) if data_format == "channels_last" else x

                x = np.pad(x, ((0, 0), (padding[0], padding[0]), (padding[1], padding[1])))

                if data_format == "channels_last":
                    x = np.moveaxis(x, 0, -1)

            elif x.ndim == 2:  # A single grayscale image
                x = np.pad(x, ((padding[0], padding[0]), (padding[1], padding[1])))
            else:
                raise ValueError("Image must be 2D or 3D.")
        else:
            if x.ndim == 4:  # A batch of multichannel images

                # If the data format is channels_last, then we need to convert the image to channels_first
                x = np.moveaxis(x, -1, 1) if data_format == "channels_last" else x

                # The first dimension is the batch dimension
                x = np.pad(x, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])))

                if data_format == "channels_last":
                    x = np.moveaxis(x, 1, -1)

            elif x.ndim == 3:  # Batch of grayscale images
                x = np.pad(x, ((0, 0), (padding[0], padding[0]), (padding[1], padding[1])))
            else:
                raise ValueError("Image must be 2D or 3D.")

        return x

    def __call__(self, image: np.ndarray, kernel: np.ndarray, **kwargs) -> np.ndarray:
        return self.convolve(image, kernel, **kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}({self.kernel_size}, {self.padding}, {self.stride})"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the Sequential model class."""


from collections import deque
from copy import deepcopy
import numpy as np
from typing import List, Dict, Tuple
from tqdm import tqdm
import pickle


from . import Model
from dlfs.layers import Layer, Conv2D
from dlfs.optimizers import Optimizer, get_optimizer
from dlfs.losses import LossFunction, get_loss_function
from dlfs.metrics import Metric, get_metric
from dlfs.preprocessing import train_test_split


class Sequential(Model):
    """A sequential model is a linear stack of layers.

    Args:
        layers (List[Layer]): A list of layers.
        name (str): The name of the model.

    Attributes:
        layers (List[Layer]): A list of layers.
        name (str): The name of the model.
        optimizer (Optimizer): The optimizer used to optimize the model.
        loss (LossFunction): The loss function used to compute the loss.
        metrics (Dict[str, Union[Metric, LossFunction]]): A dictionary of metrics.
        __trainable (bool): Whether the model is trainable.

    Usage:
        >>> from dlfs.layers import Dense

        >>> model = Sequential([Dense(10, activation='relu'), Dense(1)])

        >>> model.summary()

        >>> model.compile(optimizer='sgd', loss='mse', metrics=['mae'])

        >>> model.fit(x_train, y_train, epochs=10)
    """

    layers: List[Layer]
    name: str
    loss: LossFunction or None
    optimizer: Optimizer or None
    metrics: Dict[str, Metric] or None
    trainable: bool

    def __init__(self, layers: List[Layer] = None, name: str = "Sequential"):

        super().__init__(name=name)

        # check that the first layers has set the input shape
        if layers and layers[0].input_shape is None:
            raise ValueError("The first layer must have an input shape")
        self.layers = []
        if layers is not None:
            for layer in layers:
                self.add(layer)
        self.name = name
        self.loss = None
        self.optimizer = None
        self.metrics = None
        self.__trainable = True
        self.__counter = 1

    @property
    def trainable(self):
        return self.__trainable

    @trainable.setter
    def trainable(self, value: bool):
        self.__trainable = value
        for layer in self.layers:
            layer.trainable = value

    def add(self, layer: Layer):
        """Adds a layer to the model and initializes it.

        Args:
            layer (Layer): the layer to add
        """

        # check that the first layers has set the input shape
        if not self.layers and layer.input_shape is None:
            raise ValueError("The first layer must have an input shape")

        # initialize the layer
        if not layer.initialized:
            if layer.input_shape is None:
                layer.initialize(input_shape=self.layers[-1].output_shape)
            else:
                layer.initialize(input_shape=layer.input_shape)

        # each layer must have an unique name
        layer.name = f"{layer.name}{self.__counter}"
        self.__counter += 1

        self.layers.append(layer)

    def concatenate(self, model: 'Sequential'):
        """Concatenates two models.
        
        Args:
            model (Sequential): the model to concatenate
        """
        if self.layers[-1].output_shape != model.layers[0].input_shape:
            raise ValueError("The two models must have the same input shape")
        self.layers += model.layers

    def compile(self, optimizer: Optimizer or str, loss: LossFunction or str, metrics: List[str] = None):
        """Sets the optimizer, loss function and metrics.

        Args:
            optimizer: the optimizer to use
            loss: the loss function to use
            metrics: the metrics to use. The metrics by default are the ones used by the loss function
            Allowed values are:
                - "accuracy"
        """
        self.optimizer = get_optimizer(optimizer) if isinstance(optimizer, str) else optimizer
        self.loss = get_loss_function(loss) if isinstance(loss, str) else loss
        self.metrics = {} if metrics is None else {metric: get_metric(metric) for metric in metrics}

        # check that each layer has an unique name
        names = set()
        for layer in self.layers:
            if layer.name in names:
                raise ValueError(f"The layer name {layer.name} is not unique")
            names.add(layer.name)
            self.optimizer.add_slot(layer)

    def summary(self):
        """Prints the summary of the model as a table."""
        print(f"Model: {self.name}")
        print("-" * len(self.name))
        print("Layers:")
        for layer in self.layers:
            print("-" * len(layer.summary()))
            print(f"{layer.summary()}")

        print("-" * len(self.name))
        print("Total params:", sum(layer.count_params() for layer in self.layers))
        print("Trainable params:", sum(layer.count_params() for layer in self.layers if layer.trainable))
        print("-" * len(self.name))

    def get_deltas(self, y_pred: np.ndarray, y_true: np.ndarray) -> deque:
        """Iterates over the layers and returns the deltas of the output of each layer.

        Delta is the gradient of the loss function with respect to the output of the layer (z).

        The deltas are returned in a deque, where the first element is the delta of the first layer,
        the second element is the delta of the second layer and so on.


        Args:
            y_pred: the predictions
            y_true: the true labels

        Returns:
            the deltas of the loss function (as a deque)
        """

        # initialize the deltas
        deltas = deque()

        last_layer = self.layers[-1]
        deltas.appendleft(last_layer.get_delta(self.loss.gradient(y_true, y_pred)))

        # backward pass
        for layer in reversed(self.layers[:-1]):
            d_inputs = last_layer.get_d_inputs(deltas[0])
            delta = layer.get_delta(d_inputs)
            deltas.appendleft(delta)
            last_layer = layer

        return deltas

    def train_on_batch(self, x_batch: np.ndarray, y_batch: np.ndarray) -> Dict[str, float]:
        """Trains the model on a batch of data.

        Args:
            x_batch: the batch of inputs
            y_batch: the batch of expected outputs

        Returns:
            the loss and the metrics
        """

        # forward pass
        y_pred = self.predict(x_batch, training=True)

        # backward pass: get the deltas
        deltas: deque = self.get_deltas(y_pred, y_batch)

        # backward pass: update the weights
        for i, layer in enumerate(self.layers):
            if layer.trainable:
                layer.update(self.optimizer, deltas[i])

        # compute the loss for the batch
        loss = self.loss.compute_loss(y_batch, y_pred)

        # compute the metrics for the batch
        metrics = {metric.name: metric(y_batch, y_pred) for metric in self.metrics.values()}

        return {**{'loss': loss}, **metrics}

    def __check_data(self, x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Checks that the data is valid and adds a batch dimension if needed."""

        if isinstance(self.layers[0], Conv2D) and len(x.shape) == 3:
            # add a channel dimension
            x = x[:, np.newaxis, :, :]
            # change to channels last
            x = np.moveaxis(x, 1, -1)

        if x.shape[1:] != self.layers[0].input_shape[1:]:
            raise ValueError(f"The input shape of the model is {self.layers[0].input_shape}"
                             f" but the input shape of the data is {x.shape}")

        if y.ndim == 1:
            y = y.reshape(-1, 1)

        # check if the labels are valid
        if y.shape[1:] != self.layers[-1].output_shape[1:]:
            raise ValueError(f"The output shape of the model is {self.layers[-1].output_shape}")

        return x, y

    def __update_io_shapes(self, batch_size: int):
        """Updates the input and output shapes of the layers to match the batch size.

        Args:
            batch_size: the batch size
        """
        for layer in self.layers:
            layer.input_shape = (batch_size, *layer.input_shape[1:])
            layer.output_shape = (batch_size, *layer.output_shape[1:])

    def __print_results(self,
                        metrics: Dict[str, float],
                        val_metrics: Dict[str, float] = None,
                        title: str = "Epoch",
                        progress: int = 0,
                        total: int = 0,
                        history: Dict[str, list] = None) -> None:
        """Prints the results of the training and updates the history.

        Args:
            metrics: the metrics (and loss) of the current epoch
            val_metrics: the metrics of the validation set
            title: the title of the printed results (Batch or Epoch)
            progress: the progress of the training. The number of batches processed or the number of epochs
            total: the total number of batches or epochs
            history: the history of the training to update
        """

        using_validation = val_metrics is not None
        if using_validation:
            # print the results
            print(f"{title} ({progress}/{total}) - loss: {metrics['loss']:.4f}"
                  f" - val_loss: {val_metrics['val_loss']:.4f}")
            for metric in self.metrics:
                print(f"- {metric}: {metrics[metric]:.4f} - val_{metric}: {val_metrics['val_' + metric]:.4f}")
        else:
            # print the results
            print(f"{title} ({progress}/{total}) - loss: {metrics['loss']:.4f}")
            for metric in self.metrics:
                print(f"- {metric}: {metrics[metric]:.4f}")

        if history is not None:
            for metric in metrics:
                history[metric].append(metrics[metric])
            if using_validation:
                for metric in val_metrics:
                    history[metric].append(val_metrics[metric])
        print("")

    def fit(self,
            x: np.ndarray,
            y: np.ndarray,
            epochs: int = 1,
            batch_size: int = 32,
            verbose: int = 1,
            validation_data: Tuple[np.ndarray, np.ndarray] = None,
            validation_split: float = 0.,
            shuffle: bool = True,
            initial_epoch: int = 0) -> Dict[str, List[float]]:
        """Trains the model.

        Args:
            x: the input data. The first dimension is the batch size.
            y: The expected values. The first dimension is the batch size.
            epochs: the number of epochs to train the model.
            batch_size: The number of samples in each batch.
            verbose: the verbosity mode:
                * 0 doesn't show anything.
                * 1 prints one line per batch.
                * 2 prints one line per epoch.
                * 3 shows the progress bar (per epoch).
            validation_data: the validation data. Only used if validation_split is not 0 because its function is
                to see how the model performs on the validation set.
            validation_split: the percentage of the training set to use as validation data.
            shuffle: whether to shuffle the data before each epoch.
            initial_epoch: the initial epoch.
        """
        # CHECKS:
        # check if the model has been compiled
        if self.optimizer is None or self.loss is None:
            raise ValueError("You must compile the model before training")

        # check if the data dimensions are valid
        x, y = self.__check_data(x, y)

        using_validation_data = True
        # if validation data is not provided, split the data into train and validation
        if validation_split > 0:
            if validation_data is not None:
                raise ValueError("validation_data and validation_split cannot be used together")
            x, x_val, y, y_val = train_test_split(x, y, test_size=validation_split)
        elif validation_data is not None:
            x_val, y_val = validation_data

            # check if the data dimensions are valid
            x_val, y_val = self.__check_data(x_val, y_val)
        else:
            x_val = None
            y_val = None
            using_validation_data = False

        if using_validation_data and verbose == 0:
            raise ValueError("validation_data and verbose=0 cannot be used together")

        # INITIALIZATION:
        # --------------------------------------------------

        # Update the input_shape of the layers to take into account the batch_size
        for layer in self.layers:
            layer.input_shape = (batch_size, *layer.input_shape[1:])

        # Update the output_shape of the layers to take into account the batch_size
        for layer in self.layers:
            layer.output_shape = (batch_size, *layer.output_shape[1:])

        # initialize the history
        history = {"loss": [], "val_loss": []}
        for metric in self.metrics:
            history[metric] = []
            history["val_" + metric] = []

        # set the optimizers
        for layer in self.layers:
            layer.optimizer = deepcopy(self.optimizer)

        # TRAINING:
        # --------------------------------------------------
        # create the progress bar if verbose is 0
        r = range(initial_epoch, epochs)
        range_epochs = r if verbose != 3 else tqdm(r)

        # loop over the number of epochs
        for epoch in range_epochs:
            # initialize the total loss for the epoch if verbose is 2

            epoch_metrics = {metric: 0.0 for metric in self.metrics}
            epoch_metrics['loss'] = 0.0

            val_metrics = None

            # loop over the data in batches
            total_data_used_per_epoch = 0
            generator = self.batch_generator(x, y, batch_size, shuffle)
            if verbose == 2:
                generator = tqdm(generator, total=len(x) // batch_size)

            for x_batch, y_batch in generator:

                total_data_used_per_epoch += x_batch.shape[0]

                metrics = self.train_on_batch(x_batch, y_batch)

                # update the total loss and metrics for the epoch
                for metric in epoch_metrics:
                    epoch_metrics[metric] += metrics[metric]

                if using_validation_data and verbose == 1:
                    # compute the loss and the metrics for the validation data
                    y_val_pred = self.predict(x_val)
                    val_loss = self.loss.compute_loss(y_val, y_val_pred)
                    val_metrics = {"val_" + metric: self.metrics[metric](y_val, y_val_pred)
                                   for metric in self.metrics}

                    # add val_loss to val_metrics
                    val_metrics["val_loss"] = val_loss

                # print the loss and the metrics per batch (if verbose is 1)
                if verbose == 1:
                    self.__print_results(metrics, val_metrics, "Batch",
                                         progress=total_data_used_per_epoch, total=len(x))

            # print the metrics of the epoch (if verbose is 1 or 2)
            if verbose == 1 or verbose == 2:

                # compute the average loss and metrics for the epoch
                epoch_metrics = {metric: epoch_metrics[metric] / (len(x) // batch_size) for metric in epoch_metrics}
                val_epoch_metrics = None
                if using_validation_data:
                    # compute the loss and the metrics for the validation data
                    y_val_pred = self.predict(x_val)
                    val_epoch_loss = self.loss.compute_loss(y_val, y_val_pred)
                    val_epoch_metrics = {"val_" + metric: self.metrics[metric](y_val, y_val_pred)
                                         for metric in self.metrics}

                    # add val_loss to val_metrics
                    val_epoch_metrics["val_loss"] = val_epoch_loss

                # print the metrics of the epoch and update the history
                self.__print_results(epoch_metrics, val_epoch_metrics, "Epoch",
                                     progress=epoch + 1, total=epochs, history=history)

        if verbose == 3:
            # print final results
            self.evaluate(x, y)
            if using_validation_data:
                self.evaluate(x_val, y_val, prefix="val_")
        return history

    @staticmethod
    def batch_generator(x: np.ndarray, y: np.ndarray, batch_size: int, shuffle: bool = True):
        """Generates batches of data.

        Args:
            x: the input data
            y: the labels
            batch_size: the batch size
            shuffle: whether to shuffle the data

        Yields:
            x_batch: the batch of input data
            y_batch: the batch of labels
        """
        # get the number of batches
        n_batches = len(x) // batch_size

        if shuffle:
            # shuffle the data
            indices = np.arange(len(x))
            np.random.shuffle(indices)
            x = x[indices]
            y = y[indices]

        # loop over the batches
        for i in range(0, n_batches * batch_size, batch_size):
            # get the batch data
            x_batch = x[i:i + batch_size]
            y_batch = y[i:i + batch_size]
            # yield the batch
            yield x_batch, y_batch

    def evaluate(self,
                 x: np.ndarray,
                 y: np.ndarray,
                 batch_size: int = None,
                 verbose: int = 1,
                 prefix: str = "") -> dict:
        """Evaluates the model on the given data.

        Args:
            x: the input data
            y: the labels
            batch_size: the batch size (default: len(image))
            verbose: the verbosity mode (0, 1, or 2)
            prefix: the prefix to print

        Returns:
            The average loss and the average metrics of the model on the given data
        """
        # check the data
        x, y = self.__check_data(x, y)

        # get the batch size
        batch_size = batch_size or len(x)

        # update input and output shapes of the model
        self.__update_io_shapes(batch_size)

        # initialize the metrics
        avg_loss = 0.0
        avg_metrics = {prefix + metric: 0.0 for metric in self.metrics}

        n_batches = len(x) // batch_size

        # loop over the batches
        for i in range(0, n_batches * batch_size, batch_size):
            # get the batch data
            y_pred = x[i:i + batch_size]
            y_batch = y[i:i + batch_size]

            # forward pass
            y_pred = self.predict(y_pred)

            # compute the loss for the batch
            loss = self.loss.compute_loss(y_batch, y_pred)

            # compute the metrics for the batch
            metrics = {prefix + metric: self.metrics[metric](y_batch, y_pred) for metric in self.metrics}

            # update the loss and the metrics
            avg_loss += 1/(i + 1) * (loss - avg_loss)
            for metric in self.metrics:
                avg_metrics[prefix + metric] += 1/(i + 1) * (metrics[prefix + metric] - avg_metrics[prefix + metric])

            # print the loss and the metrics per batch (if verbose is 1)
            if verbose == 2 and batch_size != len(x):
                print(f"Batch ({i // batch_size + 1}/{n_batches}) - loss: {loss}")
                print(f"\t{', '.join([prefix + f'{metric}: {metrics[prefix + metric]:.4f}' for metric in self.metrics])}")

        # print final loss and metrics
        if verbose > 0:
            print(prefix + f"loss: {avg_loss}")
            print(f"{', '.join([prefix + f'{metric}: {avg_metrics[prefix + metric]:.4f}' for metric in self.metrics])}")

        return {prefix + 'loss': avg_loss, **avg_metrics}

    def predict(self, x: np.ndarray, training: bool = False) -> np.ndarray:
        """Predicts the output of the model on the given data.

        Args:
            x : the input data
            training: whether the model is in training mode
        Returns:
            y_pred : the predictions
        """

        if not training:
            self.__update_io_shapes(len(x))

        last_input = x
        for layer in self.layers:
            last_input = layer.forward(last_input, training)
        return last_input

    def set_weights(self, weights: list, layers: list = None, only_weights: bool = False, only_biases: bool = False):
        """Sets the weights and biases of the model. The weights and biases are in the same order as the layers.

        Args:
            weights: the weights of the model (list of numpy arrays) [layer_0_weights, layer_0_bias, ...]
                The weights and biases of all the layers that have them must be included in the list
            layers: the layers of the model (list of Layer objects) which you want to set the weights of
                (default: all layers with weights)
            only_weights: whether to set only the weights. If True, the weights list must contain
                only the weights of the layers in the layers list. (default: False)
            only_biases: whether to set only the biases. If True, the weights list must contain
                only the biases of the layers in the layers list. (default: False)
        """
        layers_with_weights = layers or [layer for layer in self.layers if layer.has_weights]

        if only_weights:
            for i, w in enumerate(weights):
                layers_with_weights[i].set_weights(w)
        elif only_biases:
            for i, b in enumerate(weights):
                layers_with_weights[i].set_weights(bias=b)
        else:
            for i, layer in enumerate(layers_with_weights):
                layer.set_weights(weights[i], weights[i + len(layers_with_weights)])

    def save(self, path: str):
        """Saves the model to the given path using the pickle module.

        Args:
            path: the path to save the model
        """
        with open(path, 'wb') as f:
            pickle.dump(self, f)

    @staticmethod
    def load(path: str) -> 'Sequential':
        """Loads the model from the given path using the pickle module.

        Args:
            path: the path to load the model
        """
        with open(path, 'rb') as f:
            return pickle.load(f)

    def __str__(self):
        return f"{self.__class__.__name__}({', '.join([str(layer) for layer in self.layers])})"

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.layers)

    def __getitem__(self, index: int):
        return self.layers[index]

    def __iter__(self):
        return iter(self.layers)

    def __add__(self, other: 'Sequential'):
        return Sequential(self.layers + other.layers)

    def __iadd__(self, other: 'Sequential'):
        self.concatenate(other)

    def __copy__(self):
        return deepcopy(Sequential(self.layers))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the PatchConvolutioner class-."""

import numpy as np
from typing import Union


from .convolutioner import Convolutioner


class PatchConvolutioner(Convolutioner):
    """Convolutioner for patch-based convolutioners.

    This class implements the convolutioner interface for patch-based convolutioners. For avoid using
    for loops (which are slow), we can take advantage of how numpy arrays are stored in memory.

    This class is not ready for use yet.

    Args:
        image_size (tuple[int, int] or int): The size of the image to convolve. If an int is provided,
            it is assumed to be the size of the image along the first two dimensions.
        kernel_size (tuple[int, int] or int): The size of the kernel to convolve. If an int is provided,
            it is assumed to be the size of the kernel along the first two dimensions.
        padding (tuple[int, int] or int): The padding to apply to the image. If an int is provided,
            it is assumed to be the padding along the first two dimensions.
        stride (tuple[int, int] or int): The stride of the convolution. If an int is provided,
            it is assumed to be the stride along the first two dimensions.
        data_format (str): The data format of the image and kernel. Must be either 'channels_first' or 'channels_last'.
    """
    def __init__(self,
                 image_size: Union[int, tuple],
                 kernel_size: Union[int, tuple],
                 padding: Union[int, tuple] = (0, 0),
                 stride: Union[int, tuple] = (1, 1),
                 data_format: str = 'channels_last'):

        super().__init__(image_size, kernel_size, padding, stride, data_format)

    @staticmethod
    def convolve_grayscale(image: np.ndarray, kernel: np.ndarray, padding: Union[int, tuple] = (0, 0),
                           stride: Union[int, tuple] = (1, 1), using_batches: bool = False) -> np.ndarray:
        raise NotImplementedError

    @staticmethod
    def convolve_multichannel(image: np.ndarray, kernel: np.ndarray, padding: tuple = (0, 0),
                              stride: Union[int, tuple] = (1, 1), using_batches: bool = False) -> np.ndarray:
        pass



from distutils.core import setup

setup(
    name='dlfs',
    packages=['dlfs',
              'dlfs.activation_functions',
              'dlfs.convolutioners',
              'dlfs.layers',
              'dlfs.losses',
              'dlfs.metrics',
              'dlfs.models',
              'dlfs.optimizers',
              'dlfs.preprocessing'],

    version='0.2.3',
    license='apache-2.1',
    description='implement from scratch (using numpy arrays) a package based on tensorflow architecture which '
                'allows to build and train Fully Connected Networks and Convolutional Neural Networks (CNNs).',
    author='Pablo',
    author_email='pablete.arino@gmail.com',
    url='https://github.com/Pabloo22/Deep-Learning-from-Scratch',
    download_url='https://github.com/Pabloo22/dlfs/archive/refs/tags/v0.2.3.tar.gz',
    keywords=['machine-learning', 'deep-learning', 'backpropagation'],  # Keywords that define your package best
    install_requires=[
        'numpy',
        'tqdm',
        'patchify',
        'tensorly',
        'scikit-image'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the Linear Activation function"""

import numpy as np
from . import ActivationFunction


class Linear(ActivationFunction):
    """Linear activation function.

    This activation function is the identity function, it does not change the
    input.
    """
    def __init__(self):
        super().__init__(name='linear')

    @staticmethod
    def forward(x):
        return x

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        """Returns a tensor of the same shape as z, with all elements set to 1."""
        return np.ones_like(z)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the accuracy metric."""

import numpy as np
from .metric import Metric


class Accuracy(Metric):

    def __init__(self, name='accuracy'):
        super().__init__(name)

    @staticmethod
    def compute_metric(y_true: np.ndarray, y_pred: np.ndarray):
        """
        Compute the accuracy of the model.

        Args:
            y_true: The true labels.
            y_pred: The predicted labels.

        Returns:
            The accuracy of the model.
        """
        using_multi_label = len(y_true[0]) > 1
        y_pred = np.argmax(y_pred, axis=1) if using_multi_label else np.round(y_pred)
        y_true = np.argmax(y_true, axis=1) if using_multi_label else y_true

        # Return the accuracy of the model (multilabel or binary)
        accuracy = np.mean(y_pred == y_true)
        return accuracy




"""
The underlying concept of pooling is to reduce size and accumulate the most remarkable information by rolling the
filter over the image. This is due to reduction in the number of parameters and computations in the network.
"""
import numpy as np
from typing import Union
from .layer import Layer


class MaxPooling2D(Layer):
    """MaxPooling2D class.

    This is the most common form of pooling. It is based on the idea of extracting the
    largest number from all of the cells that covers the filter. Therefore, if a feature is present
    on the image, this type of pooling will boost its significance in the resulting activation
    map.
    """

    def get_d_inputs(self, delta: np.ndarray) -> np.ndarray:
        pass

    def __init__(self,
                 pool_size: Union[int, tuple] = 2,
                 stride: Union[int, tuple] = 2,
                 padding: Union[int, tuple] = 0,
                 name="MaxPool2D",
                 activation: str = None,
                 input_shape: tuple = None, ):
        """
        MaxPool2D class constructor.
        Args:
            pool_size: size of the pooling window
            stride: stride of the pooling window
            padding: if True, add padding to the input
            name: name of the layer
            activation: activation function of the layer
            input_shape: shape of the input tensor
        """
        input_shape = (None, *input_shape) if input_shape is not None else None
        pool_size = pool_size if isinstance(pool_size, tuple) else (pool_size, pool_size)
        stride = stride if isinstance(stride, tuple) else (stride, stride)
        padding = padding if isinstance(padding, tuple) else (padding, padding)

        self.pool_size = pool_size
        self.stride = stride
        self.padding = padding

        # compute output shape (batch_size, width, height, n_channels)
        output_shape = self._get_output_shape(input_shape) if input_shape is not None else None

        # call the base class constructor
        super().__init__(name=name, activation=activation, input_shape=input_shape, output_shape=output_shape)

    def _get_output_shape(self, input_shape: tuple) -> tuple:
        """
        Given the input shape, return the output shape.
        Args:
            input_shape: shape of the input tensor (n_filters, width, height)
        """

        return (input_shape[0],  # batch_size
                int((input_shape[1] - self.pool_size[0]) / self.stride[0]) + 1,  # width
                int((input_shape[2] - self.pool_size[1]) / self.stride[1]) + 1,  # height
                input_shape[3])  # n_channels

    def initialize(self, input_shape: tuple):
        """
        Initialize the layer. This method is called by the model when the model is being compiled. In this case,
        only the input shape and the output shape are set because this layer does not have weights.

        Args:
            input_shape: The input shape.
        """
        self.input_shape = input_shape
        self.output_shape = self._get_output_shape(input_shape)
        self.initialized = True

    def forward(self, x: np.ndarray, training: bool = False) -> np.ndarray:
        """
        Forward pass of the layer.

        Args:
            x: input numpy array
            training: training mode (added for compatibility with the base Layer class)
        """
        self.inputs = x
        self.outputs = np.zeros(self.output_shape)
        for i in range(self.output_shape[1]):
            for j in range(self.output_shape[2]):
                self.outputs[:, i, j, :] = np.max(x[:, i * self.stride[0]:i * self.stride[0] + self.pool_size[0],
                                                  j * self.stride[1]:j * self.stride[1] + self.pool_size[1], :],
                                                  axis=(1, 2))

        return self.outputs if self.activation is None else self.activation(self.outputs)

    def get_delta(self, last_delta: np.ndarray, dz_da: np.ndarray) -> np.ndarray:
        # TODO: Implement backward pass
        arr = np.zeros((self.output_shape))

    def set_weights(self, weights: np.ndarray = None, bias: np.ndarray = None):
        pass

    def get_dz_da(self) -> np.ndarray:
        pass

    def update(self, optimizer, delta: np.ndarray):
        pass

    def count_params(self) -> int:
        return 0

    def summary(self) -> str:
        return f"{self.name} ({self.input_shape} -> {self.output_shape})"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the functions related to the Winograd algorithm using Vandermonde matrices. The function of this file
    is to show an implementation of some functions rather than to provide a full implementation of the Winograd
    algorithm."""

# Source: On Improving the Numerical Stability of Winograd Convolutions
# By: Kevin Vincent and Kevin J. Stephano and Michael A. Frumkin and Boris Ginsburg and Julien Demouth
# (https://openreview.net/pdf?id=H1ZaRZVKg)

import numpy as np
from typing import Union, List, Tuple
import tensorly

np.set_printoptions(formatter={
    'float': lambda x: "{0:0.3f}".format(x)
})


def vandermonde_matrix(b: int, points: List[Tuple[Union[int, float]]]) -> np.ndarray:
    """
    This is equivalent to V_{axb} on the document.
    This function returns a trimmed Vandermonde Matrix of size len(points) image b

        f_0^0 * g_0^b-1       f_0^1 * g_0^b-2       ... f_0^b-1 * g_0^0
        .
        .
        f_a-1^0 * g_a-1^b-1   f_a-1^1 * g_a-1^b-2   ... f_a-1^b-1 * g_a-1^0

    Arguments:
        b (int) : One of the dimensions of the matrix
        points (list[tuple[Union[int, float]]]) : List of homogeneous coordinates
    Returns:
        V (np.ndarray[Union[int, float]]) : The trimmed Vandermonde matrix
    """
    return np.array([[i[0] ** j * i[1] ** (b - j - 1) for j in range(b)] for i in points])


def gen_points(num_points: int) -> List[Tuple[float, int]]:
    """
    This function generates homogeneous coordinates, starting from (0, 1), then, (1/2, 1), then (-1/2, 1),
    and the infinite point (0, 1).

        [(f_0, g_0), ..., (f_a-1, g_a-1)]

    Args:
        num_points (int): The desired number of points

    Returns:
        list[tuple[Union[int, float]]]: List of homogeneous coordinates
    """
    points = [(0, 1)]
    if num_points % 2:
        points.extend([(y / 2, 1) for x in range(1, num_points // 2) for y in (x, -x)])
        points.extend([(num_points // 2 / 2, 1), (1, 0)])
    else:
        points.extend([(y / 2, 1) for x in range(1, num_points // 2) for y in (x, -x)])
        points.append((1, 0))
    return points


def winograd_get_matrices(m: int, n: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    This is equivalent to F(m, n) on the paper.
    This function returns the matrices necessary for calculating the Winograd convolution given the dimension of the filter and the dimension of the output. With those, we can infer the size of the input.
    The usage of these matrices depends on the dimension of the convolution. If we denote image as the signal and w as the filter:
    - If it is 1D: F(m, n) = image * w = Y.T @ [(X.T @ image) * (W @ w)]
    - If it is 2D: F(m image i, n image j) = image * w = Y.T @ [(X.T @ image @ X') * (W @ w @ W'.T)] @ Y'
    In the latter case, if m = i and n = j, then Y = Y', X = X', W = W'. If not, then you would have to calculate both F(m, n) to get Y, X, W and F(i, j) to get Y', X', W'.

    Args:
        m (int): The size of the output
        n (int): The size of the filter

    Returns:
        A tuple with the aforementioned Y.T, X, W in this order.
    """
    num_points = m + n - 1
    points = gen_points(num_points)
    Y = vandermonde_matrix(m, points)
    X = np.linalg.inv(vandermonde_matrix(num_points, points))
    W = vandermonde_matrix(n, points)
    return Y.T, X, W


def winograd_chunk_2D(chunk, filter, YT, X1, W1, Y, X2, W2):
    return YT @ ((X1.T @ chunk @ X2) * (W1 @ filter @ W2.T)) @ Y.T


def winograd_chunk_3D(chunk, filter, winograd_matrices):
    part1 = tensorly.tenalg.multi_mode_dot(chunk, [t[1] for t in winograd_matrices], list(range(chunk.ndim)))
    part2 = tensorly.tenalg.multi_mode_dot(filter, [t[2] for t in winograd_matrices], list(range(filter.ndim)))
    part3 = part1 * part2
    return tensorly.tenalg.multi_mode_dot(part3, [t[0] for t in winograd_matrices], list(range(part3.ndim)))


def winograd_convolution(source, filter):
    if not source.ndim == filter.ndim:
        raise IndexError
    already_calculated = []
    winograd_matrices = []
    if source.ndim == 1:
        out_dim = source.shape[0] - filter.shape[0] + 1
        YT, X, W = winograd_get_matrices(out_dim, filter.shape[0])
        out = YT @ ((X.T @ source @ X) * (W @ filter))
    elif source.ndim == 2:
        out_dim1 = source.shape[0] - filter.shape[0] + 1
        YT, X1, W1 = winograd_get_matrices(out_dim1, filter.shape[0])
        out_dim2 = source.shape[1] - filter.shape[1] + 1
        if out_dim1 != out_dim2 or filter.shape[0] != filter.shape[1]:
            Y, X2, W2 = winograd_get_matrices(out_dim2, filter.shape[1])
        else:
            Y, X2, W2 = YT, X1, W1
        # out = YT @ ((X1.T @ source @ X2) * (W1 @ filter @ W2.T)) @ Y.T
    else:
        for i, j in zip(source.shape, filter.shape):
            out_dim = i - j + 1
            if not (out_dim, j) in already_calculated:
                already_calculated.append((out_dim, j))
                winograd_matrices.append(winograd_get_matrices(out_dim, j))
            else:
                winograd_matrices.append(winograd_matrices[already_calculated.index((out_dim, j))])

        # part1 = tensorly.tenalg.multi_mode_dot(source, [t[1] for t in winograd_matrices], list(range(source.ndim)))
        # part2 = tensorly.tenalg.multi_mode_dot(filter, [t[2] for t in winograd_matrices], list(range(filter.ndim)))
        # part3 = part1 * part2
        # out = tensorly.tenalg.multi_mode_dot(part3, [t[0] for t in winograd_matrices], list(range(part3.ndim)))
    #return out


def winograd_convolution_with_blocks(source, filter, blocksize=None):
    if not source.ndim == filter.ndim:
        raise IndexError
    already_calculated = []
    winograd_matrices = []
    if source.ndim == 2:
        if blocksize == None:
            num_of_passes = (source.shape[0] - filter.size[0], source.shape[1] - filter.size[1])
            shape1 = -1
            shape2 = -1
            for i in range(source.shape[0] + 1, source.shape[0]):
                if num_of_passes[0] % shape1 == 0:
                    shape1 = i
                    break
            for i in range(source.shape[1] + 1, source.shape[1]):
                if num_of_passes[1] % shape2 == 0:
                    shape2 = i
                    break
            blocksize = np.array((shape1, shape2))
        num_of_blocks = np.array((source.shape[0]/blocksize[0], source.shape[1]/blocksize[1]))
        if np.prod(num_of_blocks).dtype != int:
            raise ValueError
        block_output_size = blocksize + 1 - filter.size
        output_size = block_output_size * num_of_blocks

if __name__ == "__main__":
    '''Y, X, W = winograd_get_matrices(3, 3)
    print('Y:\n', Y)
    print('X:\n', X)
    print('W:\n', W)'''
    test_image = np.arange(4 * 4).reshape(4, 4)
    test_filter = np.arange(2 * 2).reshape(2, 2)
    print(winograd_convolution(test_image, test_filter))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the dropout layer."""

import numpy as np
from .layer import Layer
from dlfs.optimizers import Optimizer


class Dropout(Layer):
    """Applies Dropout to the input.

    The dropout layer is a regularization technique that is used to prevent overfitting. It randomly sets a fraction
    of the input units to 0 at each update during training time using a mask, and rescales the remaining units by 1/
    (1-p) during training, where p is the dropout rate. Note that this regularization technique only applies to the
    training phase, when using the `fit` or `train_on_batch` methods.

    Args:
        p (float): The dropout rate.
        seed (int): The seed for the random number generator.
        name (str): The name of the layer.

    """

    def __init__(self, p: float, seed: int = None, name: str = "Dropout"):

        # check if p is in range [0, 1]
        if not 0 <= p <= 1:
            raise ValueError("p must be in range [0, 1]")

        super(Dropout, self).__init__(name=name, has_weights=False)

        self.prob_keep = 1 - p
        self.random_state = np.random.RandomState(seed)
        self.mask = None

    def initialize(self, input_shape: tuple):
        """Initialize input shape and output shape.

        Args:
            input_shape: shape of input data
        """
        self.input_shape = input_shape
        self.output_shape = input_shape
        self.initialized = True

    def forward(self, x: np.ndarray, training: bool = True):
        """Forward pass of the layer.

        Args:
            x: input to the layer.
            training: whether the layer is in training mode.

        Returns:
            output of the layer.
        """
        if training:
            self.mask = self.random_state.binomial(n=1, p=self.prob_keep, size=x.shape).astype(np.float32)
            return x * self.mask
        else:
            return x

    def get_d_inputs(self, delta: np.ndarray) -> np.ndarray:
        """Returns the derivative of the cost function with respect to the input of the layer.

        Args:
            delta: derivative of the cost function with respect to the output of the layer.

        Returns:
            derivative of the cost function with respect to the input of the layer.
        """
        return delta * self.mask

    def summary(self):
        """Returns a string summary of the layer."""
        return f"{self.name} (p={1 - self.prob_keep:.2f})"

    def update(self, optimizer: Optimizer, delta: np.ndarray):
        """Implemented for compatibility with the Layer interface."""

    def count_params(self) -> int:
        """Returns the number of trainable parameters in the layer."""
        return 0

    def set_weights(self, weights: np.ndarray = None, bias: np.ndarray = None):
        """Implemented for compatibility with the Layer interface."""
        raise NotImplementedError("Dropout layer has no weights")
